@@ -11,7 +11,15 @@
 # much nicer, huh?
 #
 # pylint: disable=unused-import
+from contextvars import ContextVar
+
 from contextvars_extras.context_var_descriptor import ContextVarDescriptor
 from contextvars_extras.context_var_ext import ContextVarExt
 from contextvars_extras.context_vars_registry import ContextVarsRegistry
-from contextvars_extras.sentinel import Missing
+
+__all__ = [
+    "ContextVar",
+    "ContextVarDescriptor",
+    "ContextVarExt",
+    "ContextVarsRegistry",
+]
