@@ -1,24 +1,48 @@
-from __future__ import annotations
-
+import abc
 import threading
-from collections.abc import ItemsView, KeysView, MutableMapping, ValuesView
-from contextlib import contextmanager
-from contextvars import ContextVar, Token
-from typing import Dict, List, Tuple, get_type_hints
+from collections.abc import ItemsView, KeysView, ValuesView
+from contextlib import ExitStack
+from contextvars import ContextVar
+from types import FunctionType, MethodType
+from typing import Any, ClassVar, Dict, Iterable, MutableMapping, Tuple, get_type_hints
+
+from sentinel_value import sentinel
 
 from contextvars_extras.context_var_descriptor import ContextVarDescriptor
-from contextvars_extras.context_var_ext import ContextVarValueDeleted
+from contextvars_extras.context_var_ext import DELETED, NO_DEFAULT, ContextVarExt, Token
 from contextvars_extras.internal_utils import ExceptionDocstringMixin
-from contextvars_extras.sentinel import Missing
 
 
-class ContextVarsRegistry(MutableMapping):
+class EmptySlotsMeta(abc.ABCMeta):
+    """Automatically add empty __slots__ to all registry classes.
+
+    This metaclass adds empty :ref:`slots` to :class:`ContextVarsRegistry` and all its subclasses,
+    which means that you can't set any attributes on a registry instance.
+
+    Why?
+    Because a registry doesn't have any real attributes.
+    It only acts as a proxy, forwarding operations to context variables
+    (which are hosted in the registry class, not in the instance).
+
+    Setting regular (non-context-variable) attributes on an instance would almost
+    always lead to nasty race conditions (bugs that occur in production, but not in tests).
+
+    To avoid that, we set ``__slots__ = tuple()`` for all registry classes,
+    thus ensuring that all the state is stored in context variables.
+    """
+
+    def __new__(cls, name, bases, attrs):  # noqa: D102
+        attrs.setdefault("__slots__", tuple())
+        return super().__new__(cls, name, bases, attrs)
+
+
+class ContextVarsRegistry(MutableMapping[str, Any], metaclass=EmptySlotsMeta):
     """A collection of ContextVar() objects, with nice @property-like way to access them."""
 
-    _registry_auto_create_vars: bool = True
-    """Automatically create ContextVar() objects when setting attributes?
+    _registry_allocate_on_setattr: ClassVar[bool] = True
+    """Automatically create new context variables when setting attributes?
 
-    If set to True (default), missing ContextVar() objects are automatically created when
+    If set to True (default), missing ContextVar() objects are dynamically allocated when
     setting attributes. That is, you can define an empty class, and then set arbitrary attributes:
 
         >>> class CurrentVars(ContextVarsRegistry):
@@ -31,7 +55,7 @@ class ContextVarsRegistry(MutableMapping):
     However, if you find this behavior weak, you may disable it, like this:
 
         >>> class CurrentVars(ContextVarsRegistry):
-        ...     _registry_auto_create_vars = False
+        ...     _registry_allocate_on_setattr = False
         ...     locale: str = 'en'
 
         >>> current = CurrentVars()
@@ -39,7 +63,7 @@ class ContextVarsRegistry(MutableMapping):
         AttributeError: ...
     """
 
-    _registry_descriptors: Dict[str, ContextVarDescriptor]
+    _registry_var_descriptors: ClassVar[Dict[str, ContextVarDescriptor]]
     """A dictionary of all context vars in the registry.
 
     Keys are attribute names, and values are instances of :class:`ContextVarDescriptor`.
@@ -51,7 +75,7 @@ class ContextVarsRegistry(MutableMapping):
     It is just kept for the convenience, and maybe small performance improvements.
     """
 
-    _registry_var_create_lock: threading.RLock
+    _registry_var_allocate_lock: ClassVar[threading.RLock]
     """A lock that protects against race conditions during cration of new ContextVar() objects.
 
     ContextVar() objects are lazily created on 1st attempt to set an attribute.
@@ -62,7 +86,6 @@ class ContextVarsRegistry(MutableMapping):
     So this lock is needed to ensure that ContextVar() object is created only once.
     """
 
-    @contextmanager
     def __call__(self, **attr_names_and_values):
         """Set attributes temporarily, using context manager (the ``with`` statement in Python).
 
@@ -112,28 +135,14 @@ class ContextVarsRegistry(MutableMapping):
                 ... finally:
                 ...     current.timezone = saved_timezone
         """
-        saved_state: List[Tuple[ContextVarDescriptor, Token]] = []
-        try:
-            # Set context variables, saving their state.
-            for attr_name, value in attr_names_and_values.items():
-                # This is almost like __setattr__(), except that it also saves a Token() object,
-                # that allows to restore the previous value later (via ContextVar.reset() method).
-                descriptor = self.__before_set__ensure_initialized(attr_name, value)
-                reset_token = descriptor.set(value)  # calls ContextVar.set() method
-                saved_state.append((descriptor, reset_token))
-
-            # execute code inside the ``with: ...`` block
-            yield
-        finally:
-            # Restore context variables.
-            for (descriptor, reset_token) in saved_state:
-                descriptor.reset(reset_token)  # calls ContextVar.reset() method
+        return _OverrideRegistryAttrsTemporarily(self, **attr_names_and_values)
 
     def __init_subclass__(cls):
         cls.__ensure_subclassed_properly()
-        cls._registry_descriptors = {}
-        cls._registry_var_create_lock = threading.RLock()
-        cls.__init_type_hinted_class_attrs_as_descriptors()
+        cls._registry_var_descriptors = {}
+        cls._registry_var_allocate_lock = threading.RLock()
+        cls.__convert_attrs_to_var_descriptors()
+        cls.__init_var_allocation_on_setattr()
         super().__init_subclass__()
 
     @classmethod
@@ -141,81 +150,142 @@ class ContextVarsRegistry(MutableMapping):
         if ContextVarsRegistry not in cls.__bases__:
             raise RegistryInheritanceError
 
+    @classmethod
+    def __init_var_allocation_on_setattr(cls):
+        if not cls._registry_allocate_on_setattr:
+            return
+
+        # When _registry_allocate_on_setattr=True, we extend class with extra __setattr__() method,
+        # that dynamically allocates ContextVarDescriptor() object if attribute is missing.
+        #
+        # Yeah, but why not just define a usual __setattr__() method?
+        # Why do we need to generate the method dynamically?
+        #
+        # For 2 reasons:
+        #
+        # 1. Better static code analysis: when mypy sees `def __setattr__` in a class,
+        #    it skips checks for undefined attributes. So adding the method dynamically
+        #    allows to fool mypy and thus have better type checking.
+        #
+        # 2. Better performance: native object.__setattr__ is faster than a Python method.
+        #    The overhead is small, but may become noticeable if you set variables often.
+        #    So we can avoid this overhead by NOT overriding the built-in __setattr__() method.
+        __setattr__ = cls.__setattr__
+
+        assert (
+            __setattr__ is object.__setattr__
+        ), "Customizing __setattr__() is not allowed (because then super() won't work correctly)"
+
+        def _ContextVarsRegistry__setattr__(self, attr_name, value):
+            if not hasattr(cls, attr_name):
+                cls.__before_set__ensure_allocated(attr_name, value)
+            __setattr__(self, attr_name, value)
+
+        cls.__setattr__ = _ContextVarsRegistry__setattr__  # type: ignore[assignment]
+
     # There is a bug in Pylint that gives false-positive warnings for classmethods below.
     # So, I have to mask that warning completely and wait until the bug in Pylint is fied.
     # pylint: disable=unused-private-member
 
     @classmethod
-    def __init_type_hinted_class_attrs_as_descriptors(cls):
-        hinted_attrs = get_type_hints(cls)
-        for attr_name in hinted_attrs:
-            cls.__init_class_attr_as_descriptor(attr_name)
+    def __convert_attrs_to_var_descriptors(cls):
+        for attr_name, type_hint, attr_value in _get_attr_type_hints_and_values(cls):
+            # Skip already initialized descriptors.
+            #
+            # This is needed to be able to do just this:
+            #    class CurrentVars(ContextVarsRegistry):
+            #        timezone = ContextVarDesciptor(default="UTC")
+            #
+            # So here we adopt such already existing descriptors.
+            if isinstance(attr_value, ContextVarDescriptor):
+                cls._registry_var_descriptors[attr_name] = attr_value
+                continue
+
+            # For other attributes, we may convert them to ContextVarDescriptor
+            # (the decision depends on the type hint or the attribute value).
+            if cls.__should_convert_to_descriptor(attr_name, type_hint, attr_value):
+                cls.__allocate_var_descriptor(attr_name)
+
+    @staticmethod
+    def __should_convert_to_descriptor(attr_name: str, type_hint: Any, attr_value: Any) -> bool:
+        # Have type hint?
+        # Then the decision is simple: convert all attributes that are not marked with ClassVar.
+        if type_hint is not _NO_TYPE_HINT:
+            return not _is_class_var(type_hint)
+
+        # No type hint?
+        # Then things become implicit and tricky...
+        # Here we have to use some common sense to guess which values should become context vars.
+        #
+        # The rules, roughly, are:
+        #  - skip __special__ attrs
+        #  - skip methods
+        #  - skip @property
+        #  - convert all other objects to ContextVarDescriptor
+
+        # Skip special attributes, like __doc__ and __module__
+        if _is_special_attr(attr_name):
+            return False
+
+        # Skip methods, but except for lambdas.
+        # Lambdas are converted to context variables,
+        # because usually you want to use lambdas as values, not methods.
+        if _is_method(attr_value) and not _is_lambda(attr_value):
+            return False
+
+        # Skip @property and other kinds of descriptors.
+        # ...except for lambdas again, because lambdas are technically descriptors,
+        # as they implement __get__() that converts lambda to an instance-bound method.
+        if _is_descriptor(attr_value) and not _is_lambda(attr_value):
+            return False
+
+        return True
 
     @classmethod
-    def __init_class_attr_as_descriptor(cls, attr_name):
-        with cls._registry_var_create_lock:
-            if attr_name.startswith("_registry_"):
-                return
+    def __allocate_var_descriptor(cls, attr_name):
+        with cls._registry_var_allocate_lock:
+            assert attr_name not in cls._registry_var_descriptors
 
-            assert attr_name not in cls._registry_descriptors
+            value = getattr(cls, attr_name, NO_DEFAULT)
+            assert not isinstance(value, (ContextVar, ContextVarExt, ContextVarDescriptor))
 
-            value = getattr(cls, attr_name, Missing)
-            assert not isinstance(value, (ContextVar, ContextVarDescriptor))
-
-            descriptor = ContextVarDescriptor(default=value)
+            descriptor: ContextVarDescriptor = ContextVarDescriptor(default=value)
             descriptor.__set_name__(cls, attr_name)
             setattr(cls, attr_name, descriptor)
-            cls._registry_descriptors[attr_name] = descriptor
+            cls._registry_var_descriptors[attr_name] = descriptor
 
     def __init__(self):
         self.__ensure_subclassed_properly()
         super().__init__()
 
-    def __setattr__(self, attr_name, value):
-        self.__before_set__ensure_initialized(attr_name, value)
-        super().__setattr__(attr_name, value)
-
     @classmethod
-    def __before_set__ensure_initialized(cls, attr_name, value) -> ContextVarDescriptor:
+    def __before_set__ensure_allocated(cls, attr_name, value) -> ContextVarDescriptor:
         try:
-            return cls._registry_descriptors[attr_name]
+            return cls._registry_var_descriptors[attr_name]
         except KeyError:
-            cls.__before_set__ensure_not_starts_with_special_registry_prefix(attr_name, value)
-            cls.__before_set__initialize_attr_as_context_var_descriptor(attr_name, value)
-
-            return cls._registry_descriptors[attr_name]
+            cls.__before_set__allocate_var_descriptor(attr_name, value)
+            return cls._registry_var_descriptors[attr_name]
 
     @classmethod
-    def __before_set__ensure_not_starts_with_special_registry_prefix(cls, attr_name, value):
-        if attr_name.startswith("_registry_"):
-            raise ReservedAttributeError.format(
+    def __before_set__allocate_var_descriptor(cls, attr_name, value):
+        assert cls._registry_allocate_on_setattr
+        assert not hasattr(cls, attr_name)
+        assert attr_name not in cls._registry_var_descriptors
+
+        if _is_annotated_with_class_var(cls, attr_name):
+            raise SetClassVarAttributeError.format(
                 class_name=cls.__name__,
                 attr_name=attr_name,
-                attr_type=type(value).__name__,
-                attr_value=value,
             )
 
-    @classmethod
-    def __before_set__initialize_attr_as_context_var_descriptor(cls, attr_name, value):
-        assert (
-            attr_name not in cls._registry_descriptors
-        ), "This method should not be called when attribute is already initialized as ContextVar"
-
-        if not cls._registry_auto_create_vars:
-            raise UndeclaredAttributeError.format(
-                class_name=cls.__name__,
-                attr_name=attr_name,
-                attr_type=type(value).__name__,
-            )
-
-        cls.__init_class_attr_as_descriptor(attr_name)
+        cls.__allocate_var_descriptor(attr_name)
 
     # collections.abc.MutableMapping implementation methods
 
     @classmethod
-    def _asdict(cls) -> dict:
+    def _asdict(cls) -> Dict[str, Any]:
         out = {}
-        for key, ctx_var in cls._registry_descriptors.items():
+        for key, ctx_var in cls._registry_var_descriptors.items():
             try:
                 out[key] = ctx_var.get()
             except LookupError:
@@ -286,7 +356,7 @@ class ContextVarsRegistry(MutableMapping):
 
     @classmethod
     def __getitem__(cls, key):
-        ctx_var = cls._registry_descriptors[key]
+        ctx_var = cls._registry_var_descriptors[key]
         try:
             return ctx_var.get()
         except LookupError as err:
@@ -294,24 +364,106 @@ class ContextVarsRegistry(MutableMapping):
 
     @classmethod
     def __setitem__(cls, key, value):
-        ctx_var = cls.__before_set__ensure_initialized(key, value)
+        ctx_var = cls.__before_set__ensure_allocated(key, value)
         ctx_var.set(value)
 
     def __delitem__(self, key):
-        ctx_var = self.__before_set__ensure_initialized(key, None)
-        ctx_var.__delete__(self)
+        ctx_var = self.__before_set__ensure_allocated(key, None)
+
+        if not ctx_var.is_set():
+            raise KeyError(key)
+
+        ctx_var.delete()
 
 
-def save_context_vars_registry(registry: ContextVarsRegistry) -> dict:
+_NO_ATTR_VALUE = sentinel("_NO_VALUE")
+_NO_TYPE_HINT = sentinel("_NO_TYPE_HINT")
+
+
+def _get_attr_type_hints_and_values(cls: object) -> Iterable[Tuple[str, Any, Any]]:
+    type_hints = get_type_hints(cls)
+    cls_attrs = vars(cls)
+
+    for attr_name, type_hint in type_hints.items():
+        attr_value = cls_attrs.get(attr_name, _NO_ATTR_VALUE)
+        yield (attr_name, type_hint, attr_value)
+
+    for attr_name, attr_value in cls_attrs.items():
+        if attr_name in type_hints:
+            continue
+        yield (attr_name, _NO_TYPE_HINT, attr_value)
+
+
+def _is_annotated_with_class_var(cls: type, attr_name: str) -> bool:
+    type_hints = get_type_hints(cls)
+    return (attr_name in type_hints) and _is_class_var(type_hints[attr_name])
+
+
+def _is_class_var(type_hint: object) -> bool:
+    origin = getattr(type_hint, "__origin__", None)
+    return origin is ClassVar
+
+
+def _is_method(obj: object) -> bool:
+    return isinstance(obj, (FunctionType, MethodType))
+
+
+def _is_descriptor(obj: object) -> bool:
+    return hasattr(obj, "__get__")
+
+
+def _is_special_attr(name: str) -> bool:
+    return name.startswith("__") and name.endswith("__")
+
+
+def _is_lambda(obj: object) -> bool:
+    return isinstance(obj, FunctionType) and (obj.__name__ == "<lambda>")
+
+
+class _OverrideRegistryAttrsTemporarily(ExitStack):
+    registry: ContextVarsRegistry
+    attr_names_and_values: Dict[str, Any]
+
+    def __init__(self, registry: ContextVarsRegistry, **attr_names_and_values: Dict[str, Any]):
+        self.registry = registry
+        self.attr_names_and_values = attr_names_and_values
+        super().__init__()
+
+    def __enter__(self):
+        registry = self.registry
+        registry_class = registry.__class__
+
+        for attr_name, new_value in self.attr_names_and_values.items():
+            descriptor = getattr(registry_class, attr_name, None)
+            if isinstance(descriptor, ContextVarDescriptor):
+                reset_token: Token = descriptor.set(new_value)
+                self.callback(descriptor.reset, reset_token)
+            else:
+                old_value = getattr(registry, attr_name, _NO_ATTR_VALUE)
+                setattr(registry, attr_name, new_value)
+                if old_value is _NO_ATTR_VALUE:
+                    self.callback(delattr, registry, attr_name)
+                else:
+                    self.callback(setattr, registry, attr_name, old_value)
+
+
+def save_context_vars_registry(
+    registry: ContextVarsRegistry,
+) -> Dict[str, Any]:
     """Dump variables from ContextVarsRegistry as dict.
 
     The resulting dict can be used as argument to :func:`restore_context_vars_registry`.
     """
     # pylint: disable=protected-access
-    return {key: descriptor.get_raw() for key, descriptor in registry._registry_descriptors.items()}
+    return {
+        key: descriptor.get_raw() for key, descriptor in registry._registry_var_descriptors.items()
+    }
 
 
-def restore_context_vars_registry(registry: ContextVarsRegistry, saved_registry_state: dict):
+def restore_context_vars_registry(
+    registry: ContextVarsRegistry,
+    saved_registry_state: Dict[str, Any],
+):
     """Restore ContextVarsRegistry state from dict.
 
     The :func:`restore_context_vars_registry` restores state of :class:`ContextVarsRegistry`,
@@ -361,7 +513,7 @@ def restore_context_vars_registry(registry: ContextVarsRegistry, saved_registry_
     There is still a couple of differences:
 
       1. :func:`save_registry_state` and :func:`restore_registry_state` can handle special cases,
-         like ``ContextVarValueDeleted`` tokens, or lazy initializers.
+         like ``DELETED`` tokens, or lazy initializers.
 
       2. :class:`collections.abc.MutableMapping` methods are slow, while
          :func:`save_registry_state` and :func:`restore_registry_state` are faster,
@@ -382,8 +534,8 @@ def restore_context_vars_registry(registry: ContextVarsRegistry, saved_registry_
     get_saved_value = saved_registry_state.get
 
     # pylint: disable=protected-access
-    for key, descriptor in registry._registry_descriptors.items():
-        descriptor.set(get_saved_value(key, ContextVarValueDeleted))
+    for key, descriptor in registry._registry_var_descriptors.items():
+        descriptor.context_var.set(get_saved_value(key, DELETED))
 
 
 class RegistryInheritanceError(ExceptionDocstringMixin, TypeError):
@@ -429,51 +581,22 @@ class RegistryInheritanceError(ExceptionDocstringMixin, TypeError):
     """
 
 
-class ReservedAttributeError(ExceptionDocstringMixin, AttributeError):
-    """Can't set attribute: {class_name}.{attr_name} because of the special '_registry_' prefix.
+class SetClassVarAttributeError(ExceptionDocstringMixin, AttributeError):
+    """Can't set ClassVar: '{class_name}.{attr_name}'.
 
-    This exception is raised when you try to set a special attribute::
-
-        instance.{attr_name} = {attr_value!r}
-
-    The ``_registry_*`` prefix is reserved for configuration of the :class:`ContextVarRegistry`.
-    You can't have a context variable with such name.
-
-    If you want to configure the registry class itself, you should do it when defining
-    your sub-class, like this::
+    This exception is raised when an attribute is declared as :data:`typing.ClassVar`,
+    like this::
 
         class {class_name}(ContextVarsRegistry):
-            {attr_name}: {attr_type} = {attr_value!r}
-    """
+            {attr_name}: ClassVar[...]
 
+    ...but you're trying to set it on instance level, as if it was a context variable.
 
-class UndeclaredAttributeError(ExceptionDocstringMixin, AttributeError):
-    """Can't set undeclared attribute: '{class_name}.{attr_name}'.
+    To solve the issue, you need to either:
 
-    This exception is raised when you try to set an attribute that was not declared
-    in the class {class_name} (subclass of :class:`ContextVarsRegistry`).
+    1. Remove ``ClassVar`` annotation (and thus convert the attribute to a context variable).
 
-    And, the class {class_name} is configured in a specific way
-    that disables dyanmic variable initialization::
+    2. Set the attribute off the class (not instance), like this::
 
-        class {class_name}(ContextVarsRegistry):
-            _registry_auto_create_vars = False
-
-    Because of ``_registry_auto_create_vars=False``, you can use only pre-defined variables.
-    An attempt to set any other attribute will raise this exception.
-
-    So, you have 3 options to solve the problem:
-
-    1. Add attribute to the class, like this::
-
-        class {class_name}(ContextVarsRegistry):
-            {attr_name}: {attr_type}
-
-    2. Enable dynamic initialization of new context variables, like this::
-
-        class {class_name}(ContextVarsRegistry):
-            _registry_auto_create_vars = True
-
-    3. Check the name of the attribute: '{attr_name}'.
-       Maybe there is a typo in the name?
+          {class_name}.{attr_name} = ...
     """
